"""Configuration options for behaveml functions."""

#TODO
#Add support for the user to change these.

global_config = {
    'make_movie__y_offset': 60,
    'make_movie__y_inc': 30,
    'make_movie__text_color': 'green',
    'make_movie__font_size': 36,
    'create_mosaic_video__max_mosaic_vids': 16
}
